from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractNodeMention
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.nodes import VideoNode
from ._abstract import AbstractBBox, AbstractSegment


@generate_model
@dataclass(frozen=True)
class VideoNodeMention(AbstractSegment, AbstractBBox, AbstractNodeMention):
    """
    Represents spatiotemporal volume mention for video node.

    Attributes
    --------
    node:
        The video node being mentioned.
    left:
        The left position of the spatiotemporal volume (px).
    top:
        The top position of the spatiotemporal volume (px).
    right:
         The right position of the spatiotemporal volume (px).
    bottom:
        The bottom position of the spatiotemporal volume (px).
    start:
        The start of the spatiotemporal volume (ms).
    end:
        The end of the spatiotemporal volume (ms).
    """
    node: VideoNode

    def __post_init__(self):
        if not isinstance(self.node, VideoNode):
            raise ValueError(f"Incorrect node type {type(self.node)}. Expected {VideoNode}")
        self._validate_bbox(self.node.metadata)
        self._validate_segment(self.node.metadata)



from typing import Any, Dict, Optional, Tuple, Union

from pydantic import BaseModel, ConfigDict


class MetadataModel(BaseModel):
    def to_metadata(self) -> Dict[str, Any]:
        return self.model_dump(exclude_none=True)


class FactMetadataModel(MetadataModel):
    created_time: Optional[int] = None
    modified_time: Optional[int] = None
    fact_confidence: Optional[Tuple[float]] = None
    value_confidence: Union[float, Tuple[float, ...], None] = None  # same as Optional[float, Tuple[float, ...]] (pydantic bug workaround)
    model_config = ConfigDict(extra='allow')


class DocumentMetadataModel(MetadataModel):
    title: Optional[str] = None
    file_name: Optional[str] = None
    file_type: Optional[str] = None
    size: Optional[int] = None
    created_time: Optional[int] = None
    access_time: Optional[int] = None
    modified_time: Optional[int] = None
    publication_date: Optional[int] = None
    publication_author: Optional[str] = None
    description: Optional[str] = None
    language: Optional[str] = None
    parent_uuid: Optional[str] = None
    url: Optional[str] = None
    platform: Optional[str] = None
    account: Optional[Tuple[str, ...]] = None
    access_level: Optional[str] = None
    user: Optional[str] = None
    path: Optional[str] = None
    trust_level: Optional[float] = None
    markers: Optional[Tuple[str, ...]] = None
    related_concept_id: Optional[str] = None
    preview_text: Optional[str] = None
    story: Optional[str] = None
    model_config = ConfigDict(extra='allow')



from enum import Enum
from typing import Dict, Optional, Tuple

from pydantic import BaseModel, ConfigDict


class NodeType(Enum):
    HEADER = "header"
    TEXT = "text"
    LIST = "list"
    JSON = "json"
    KEY = "key"
    TABLE = "table"
    TABLE_ROW = "row"
    IMAGE = "image"


class NodeMetadata(BaseModel):
    node_type: NodeType
    original_text: Optional[str] = None
    text_translations: Dict[str, str] = {}
    language: Optional[str] = None
    hidden: bool = False
    model_config = ConfigDict(extra='allow')


class NodeMarkup(BaseModel):
    model_config = ConfigDict(extra='allow')


class TreeDocumentContentModel(BaseModel):
    id: str
    metadata: NodeMetadata
    text: str
    nodes: Optional[Tuple['TreeDocumentContentModel', ...]] = None
    markup: NodeMarkup = NodeMarkup()


TreeDocumentContentModel.model_rebuild()



from functools import singledispatch
from typing import Dict, Iterable, Set, Tuple, Type, TypeVar, Union

from typing_extensions import Protocol

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractDomainType, AbstractFact, AbstractLinkFact, AbstractNodeMention, AbstractValue, FactStatus
from tdm.datamodel.domain import AtomValueType, ComponentValueType, PropertyType, RelationPropertyType, RelationType
from tdm.datamodel.facts import AtomValueFact, ComponentFact, ConceptFact, KBConceptValue, MentionFact, PropertyFact, \
    RelationFact, RelationPropertyFact
from tdm.utils.copy_value import copy_value


def get_metadata(document: TalismanDocument) -> Dict[str, Union[Tuple[AbstractValue, ...], AbstractValue]]:
    doc_fact = document.get_fact(document.id)
    metadata_property_facts: Set[PropertyFact] = set(document.related_facts(doc_fact, PropertyFact))

    return {property_fact.type_id.name: property_fact.target.value for property_fact in metadata_property_facts}


class MentionedFactsFactory(Protocol):
    def __call__(
            self, mention: AbstractNodeMention,
            status: FactStatus = FactStatus.NEW,
            value: Union[AbstractValue, Iterable[AbstractValue]] = ()
    ) -> Iterable[AbstractFact]:
        ...


@singledispatch
def mentioned_fact_factory(type_: AbstractDomainType) -> MentionedFactsFactory:
    raise NotImplementedError


@mentioned_fact_factory.register(AtomValueType)
def _value_fact_factory(type_: AtomValueType) -> MentionedFactsFactory:
    def create_facts(
            mention: AbstractNodeMention, status: FactStatus, value: Union[Tuple[AbstractValue, ...], AbstractValue] = ()
    ) -> Tuple[AtomValueFact, MentionFact]:
        value_fact = AtomValueFact(status, type_, value)
        return value_fact, MentionFact(status, mention, value_fact)

    return create_facts


@mentioned_fact_factory.register(PropertyType)
def _concept_fact_factory(type_: PropertyType) -> MentionedFactsFactory:
    def create_facts(
            mention: AbstractNodeMention, status: FactStatus, value: Union[Tuple[AbstractValue, ...], AbstractValue] = ()
    ) -> Tuple[ConceptFact, AtomValueFact, PropertyFact, MentionFact]:
        if status is FactStatus.APPROVED:
            concept_fact = ConceptFact(status, type_.source, KBConceptValue(value.value))
        else:
            concept_fact = ConceptFact(status, type_.source)
        value_fact = AtomValueFact(status, type_.target, value)
        return concept_fact, value_fact, PropertyFact(status, type_, concept_fact, value_fact), MentionFact(status, mention, value_fact)

    return create_facts


class LinkFactsFactory(Protocol):
    def __call__(
            self,
            source: AbstractFact,
            target: AbstractFact,
            doc: TalismanDocument,
            status: FactStatus = FactStatus.NEW
    ) -> Iterable[AbstractFact]:
        ...


@singledispatch
def link_fact_factory(type_: AbstractDomainType) -> LinkFactsFactory:
    raise NotImplementedError


@link_fact_factory.register(RelationType)
def _relation_fact_factory(type_: RelationType) -> LinkFactsFactory:
    def create_facts(
            source: AbstractFact, target: AbstractFact, doc: TalismanDocument, status: FactStatus = FactStatus.NEW
    ) -> Tuple[AbstractFact, ...]:
        return RelationFact(status, type_, source, target),

    return create_facts


_PropertyType = TypeVar('_PropertyType', bound=Type[AbstractLinkFact])


def _build_prop_fact_factory(type_: AbstractDomainType, fact_type: _PropertyType) -> LinkFactsFactory:
    def create_facts(source: AbstractFact, target: AbstractFact, doc: TalismanDocument, status=FactStatus.NEW) -> Tuple[AbstractFact, ...]:
        copied = copy_value(target, doc, status)
        return (fact_type(status, type_, source, copied[0]), *copied)

    return create_facts


@link_fact_factory.register(PropertyType)
def _(type_: PropertyType):
    return _build_prop_fact_factory(type_, PropertyFact)


@link_fact_factory.register(RelationPropertyType)
def _(type_: RelationPropertyType):
    return _build_prop_fact_factory(type_, RelationPropertyFact)


@link_fact_factory.register(ComponentValueType)
def _(type_: ComponentValueType):
    return _build_prop_fact_factory(type_, ComponentFact)



from dataclasses import replace
from typing import Sequence, TypeVar
from unittest import TestCase

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractFact, AbstractLinkFact, FactStatus
from tdm.datamodel.facts import AtomValueFact, MentionFact
from tdm.datamodel.mentions import TextNodeMention
from tdm.datamodel.nodes import TextNode
from tdm.datamodel.values import StringValue


def create_mention(node: TextNode, start, end, label) -> MentionFact:
    return MentionFact(
        FactStatus.APPROVED,
        TextNodeMention(node, start, end),
        AtomValueFact(FactStatus.APPROVED, label, StringValue(node.content[start:end]))
    )


def assert_equal_for_docs(case: TestCase, expected: Sequence[TalismanDocument], actual: Sequence[TalismanDocument]):
    case.assertEqual(len(expected), len(actual))
    for e, a in zip(expected, actual):
        assert_equal_for_facts(case, tuple(e.get_facts()), tuple(a.get_facts()))
        e = e.without_facts(e.get_facts())
        a = a.without_facts(a.get_facts())
        case.assertEqual(e, a)


def assert_equal_for_facts(case: TestCase, expected: Sequence[AbstractFact], actual: Sequence[AbstractFact]):
    case.assertEqual(len(expected), len(actual))
    case.assertSetEqual(set(map(remove_id_from_fact, expected)), set(map(remove_id_from_fact, actual)))


_Fact = TypeVar('_Fact', bound=AbstractFact)


def remove_id_from_fact(fact: _Fact) -> _Fact:
    if isinstance(fact, MentionFact):
        return replace(fact, id='', value=remove_id_from_fact(fact.value))
    if isinstance(fact, AbstractLinkFact):
        return replace(fact, id='', source=remove_id_from_fact(fact.source), target=remove_id_from_fact(fact.target))
    return replace(fact, id='')



__all__ = [
    'AudioNode', 'AudioNodeMetadata',
    'FileNode', 'FileNodeMetadata',
    'ImageNode', 'ImageNodeMetadata',
    'VideoNode', 'VideoNodeMetadata',
    'JSONNode',
    'ListNode', 'ListNodeMetadata',
    'TableCellNode', 'TableCellNodeMetadata',
    'TableNode',
    'TableRowNode',
    'KeyNode', 'TextNode', 'TextNodeMetadata'
]

from .file import AudioNode, AudioNodeMetadata, FileNode, FileNodeMetadata, ImageNode, ImageNodeMetadata, VideoNode, VideoNodeMetadata
from .structure import JSONNode, ListNode, ListNodeMetadata, TableCellNode, TableCellNodeMetadata, TableNode, TableRowNode
from .text import KeyNode, TextNode, TextNodeMetadata



from dataclasses import dataclass

from typing_extensions import Protocol


class _BBoxableNodeMetadata(Protocol):
    width: int
    height: int


@dataclass(frozen=True)
class AbstractBBox:
    """
    Abstract class for bounding box representation as [(left, top); (right, bottom)].

    Attributes
    --------
    left:
        The left position of the bounding box (px).
    top:
        The top position of the bounding box (px).
    right:
         The right position of the bounding box (px).
    bottom:
        The bottom position of the bounding box (px).
    """
    left: int
    top: int
    right: int
    bottom: int

    @property
    def width(self) -> int:
        """
        :return: bounding box width (px).
        """
        return self.right - self.left

    @property
    def height(self) -> int:
        """
        :return: bounding box height (px).
        """
        return self.bottom - self.top

    def _validate_bbox(self, metadata: _BBoxableNodeMetadata) -> None:
        if self.top < 0 or self.bottom <= self.top or self.left < 0 or self.right <= self.left:
            raise ValueError(f"Incorrect bbox [({self.left}, {self.top}); ({self.right}, {self.bottom})]")
        node_width = metadata.width
        if node_width is not None and self.right > node_width:
            raise ValueError(f"Bbox spreads out of the image (image width: {node_width}, bbox right border: {self.right})")
        node_height = metadata.height
        if node_height is not None and self.bottom > node_height:
            raise ValueError(f"Bbox spreads out of the image (image height: {node_height}, bbox bottom border: {self.bottom})")


class _SegmentableNodeMetadata(Protocol):
    duration: int


@dataclass(frozen=True)
class AbstractSegment:
    """
    Abstract class for audio segment representation as [start; end).

    Attributes
    --------
    start:
        The start of the segment (ms).
    end:
        The end of the segment (ms).
    """
    start: int
    end: int

    @property
    def duration(self) -> int:
        """
        :return: duration of the segment (ms).
        """
        return self.end - self.start

    def _validate_segment(self, metadata: _SegmentableNodeMetadata) -> None:
        if self.start < 0 or self.end <= self.start:
            raise ValueError(f"Incorrect interval [{self.start}; {self.end})")
        node_duration = metadata.duration
        if node_duration is not None and self.end > node_duration:
            raise ValueError(f"Audio segment [{self.start}; {self.end}) spreads out of the audio (duration: {node_duration})")



import dataclasses
from dataclasses import fields
from typing import Any, Dict, Generic, Optional, Type, TypeVar

from pydantic import BaseModel, ConfigDict, Field, create_model
from typing_extensions import Literal, Self

from tdm.helper import cache_result, generics_mapping, register_in_module
from .serializers import AbstractElementModel, AbstractElementSerializer, get_serializer

_Element = TypeVar('_Element')


def remove_serializers(schema: Dict[str, Any]):
    for prop in schema.get('properties', {}).values():
        prop.pop('_serializer', None)


class ElementModel(BaseModel, AbstractElementModel[_Element], Generic[_Element]):
    # TODO[pydantic]: do not include serializers into model and remove dataclass from AbstractElementSerializer
    model_config = ConfigDict(extra='forbid', json_schema_extra=remove_serializers)

    def deserialize(self, typed_id2element: Dict[type, Dict[str, Any]]) -> _Element:
        raise NotImplementedError

    @classmethod
    def serialize(cls, element: _Element) -> Self:
        kwargs = {}
        for key, value in element.__dict__.items():
            name = key[1:] if key.startswith('_') else key
            kwargs[name] = cls.model_fields[name].json_schema_extra['_serializer'].serialize(value)
        return cls(**kwargs)


def _wrap_deserializer(coll_type: type, serializer: AbstractElementSerializer):
    def deserialize(values, *args):
        return coll_type(serializer.deserialize(v, *args) for v in values)

    return deserialize


_BaseType = TypeVar('_BaseType', bound=type)


@cache_result()
def create_model_for_type(type_: Type[_Element], label: Optional[str] = None) -> Type[ElementModel[_Element]]:
    type_vars = generics_mapping(type_)

    model_fields = {}

    for field in fields(type_):
        name = field.name
        if name.startswith('_'):
            name = name[1:]
        default_value = field.default if field.default is not dataclasses.MISSING else ...

        field_type, serializer = get_serializer(field.type, type_vars)

        if hasattr(field_type, '__metadata__'):  # field type is already Annotated
            metadata = field_type.__metadata__
            if isinstance(metadata, tuple):
                metadata = metadata[0]
            discriminator = metadata.discriminator
            model_fields[name] = (field_type.__origin__, Field(default_value, discriminator=discriminator, _serializer=serializer))
        else:
            model_fields[name] = (field_type, Field(default_value, _serializer=serializer))
    if label:
        model_fields['type'] = (Literal[label], label)

    model = register_in_module(create_model(f"{type_.__name__}Model", __base__=ElementModel, **model_fields))

    def deserialize(self: model, typed_id2element: Dict[type, Dict[str, Any]]) -> type_:
        kwargs = {}
        for f, info in type_.__dataclass_fields__.items():
            if not info.init:
                continue
            n = f[1:] if f.startswith('_') else f
            if n not in self.__dict__:
                continue
            kwargs[f] = self.model_fields[n].json_schema_extra['_serializer'].deserialize(getattr(self, n), typed_id2element)
        return type_(**kwargs)

    model.deserialize = deserialize

    return model



__all__ = [
    'copy_value',
    'LinkFactsFactory', 'MentionedFactsFactory', 'get_metadata', 'link_fact_factory', 'mentioned_fact_factory',
    'get_mentions',
    'dfs'
]

from .copy_value import copy_value
from .fact_factory import LinkFactsFactory, MentionedFactsFactory, get_metadata, link_fact_factory, mentioned_fact_factory
from .mentions import get_mentions
from .node import dfs



from immutabledict import immutabledict
from pydantic import BaseModel

from tdm.helper import freeze_dict


class AbstractMarkupModel(BaseModel):
    """
    The most base class for node markup serialization
    """

    def immutabledict(self) -> immutabledict:
        return freeze_dict(self.model_dump(exclude_none=True, exclude_defaults=True))



from dataclasses import replace
from functools import singledispatch
from typing import Optional, Tuple, TypeVar

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractFact, FactStatus
from tdm.datamodel.facts import AtomValueFact, ComponentFact, CompositeValueFact, MentionFact

_Fact = TypeVar('_Fact', bound=AbstractFact)


@singledispatch
def copy_value(value: _Fact, doc: TalismanDocument, status: Optional[FactStatus]) -> Tuple[AbstractFact, ...]:
    """ Recursive copy all value information
        returns:
        - value with new id
        - mentions, components with values with new ids
    """
    raise NotImplementedError


@copy_value.register
def _(value: AtomValueFact, doc: TalismanDocument, status: Optional[FactStatus]) -> Tuple[AbstractFact, ...]:
    status = status or value.status
    new_value = replace(value, id=None, status=status)
    mentions = (replace(m, value=new_value, id=None, status=status) for m in doc.get_facts(MentionFact, filter_=lambda m: m.value == value))
    return (new_value, *mentions)


@copy_value.register
def _(value: CompositeValueFact, doc: TalismanDocument, status=Optional[FactStatus]) -> Tuple[AbstractFact, ...]:
    status = status or value.status
    new_value = replace(value, id=None, status=status)
    components = []
    for component in doc.related_facts(value, ComponentFact):
        new_component_values = copy_value(component.target, doc, status)
        new_component = ComponentFact(status, component.type_id, new_value, new_component_values[0], component.value, id=None)
        components.extend((new_component, *new_component_values))

    return (new_value, *components)



from collections import defaultdict
from functools import singledispatch
from typing import Dict, List

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractFact
from tdm.datamodel.facts import AtomValueFact, ComponentFact, CompositeValueFact, ConceptFact, MentionFact, PropertyFact, RelationFact
from tdm.datamodel.mentions import TextNodeMention
from tdm.datamodel.nodes import TextNode


@singledispatch
def get_mentions(fact: AbstractFact, document: TalismanDocument) -> Dict[TextNode, List[TextNodeMention]]:
    raise NotImplementedError


@get_mentions.register
def _(fact: AtomValueFact, document: TalismanDocument) -> Dict[TextNode, List[TextNodeMention]]:
    result = defaultdict(list)
    for mention in document.related_facts(fact, MentionFact):
        result[mention.mention.node].append(mention.mention)
    return result


@get_mentions.register
def _(fact: CompositeValueFact, document: TalismanDocument) -> Dict[TextNode, List[TextNodeMention]]:
    result = defaultdict(list)
    for component in document.related_facts(fact, ComponentFact):
        for node, mentions in get_mentions(component.target, document).items():
            result[node].extend(mentions)
    return result


@get_mentions.register
def _(fact: ConceptFact, document: TalismanDocument) -> Dict[TextNode, List[TextNodeMention]]:
    result = defaultdict(list)
    for prop in document.related_facts(fact, PropertyFact, filter_=lambda pf: pf.type_id.isIdentifying):
        prop: PropertyFact
        for node, mentions in get_mentions(prop.target, document).items():
            result[node].extend(mentions)
    return result


@get_mentions.register
def _(fact: RelationFact, document: TalismanDocument) -> Dict[TextNode, List[TextNodeMention]]:
    result = defaultdict(list)
    for concept in [fact.source, fact.target]:
        for node, mentions in get_mentions(concept, document).items():
            result[node].extend(mentions)
    return result


@get_mentions.register
def _(fact: PropertyFact, document: TalismanDocument) -> Dict[TextNode, List[TextNodeMention]]:
    result = defaultdict(list)

    for node, mentions in get_mentions(fact.target, document).items():
        result[node].extend(mentions)
    return result



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractNodeMention
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.nodes import AudioNode
from ._abstract import AbstractSegment


@generate_model
@dataclass(frozen=True)
class AudioNodeMention(AbstractSegment, AbstractNodeMention):
    """
    Represents audio segment node mention [start; end).

    Attributes
    --------
    node:
        The audio node being mentioned.
    start:
        The start of the segment (ms).
    end:
        The end of the segment (ms).
    """
    node: AudioNode

    def __post_init__(self):
        if not isinstance(self.node, AudioNode):
            raise ValueError(f"Incorrect node type {type(self.node)}. Expected {AudioNode}")
        self._validate_segment(self.node.metadata)



from dataclasses import dataclass

from typing_extensions import Self

from tdm.abstract.datamodel import AbstractNodeMention
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.nodes import ImageNode
from ._abstract import AbstractBBox


@generate_model
@dataclass(frozen=True)
class ImageNodeMention(AbstractBBox, AbstractNodeMention):
    """
    Represents bounding box image node mention.

    Attributes
    --------
    node:
        The image node being mentioned.
    left:
        The left position of the bounding box (px).
    top:
        The top position of the bounding box (px).
    right:
         The right position of the bounding box (px).
    bottom:
        The bottom position of the bounding box (px).
    """
    node: ImageNode

    def __post_init__(self):
        if not isinstance(self.node, ImageNode):
            raise ValueError(f"Incorrect node type {type(self.node)}. Expected {ImageNode}")
        self._validate_bbox(self.node.metadata)

    @classmethod
    def build(cls, node: ImageNode, left: int, top: int, width: int, height: int) -> Self:
        return cls(node, left, top, left + width, top + height)



__all__ = [
    'AudioNodeMention',
    'ImageNodeMention',
    'NodeMention',
    'TextNodeMention',
    'VideoNodeMention'
]

from .audio import AudioNodeMention
from .image import ImageNodeMention
from .node import NodeMention
from .text import TextNodeMention
from .video import VideoNodeMention
